(** Components *)

type delta = float

(** Entities *)

external c2i : 'a -> int = "%identity"

let cmp_comp a b = (c2i a) - (c2i b)

let cmp_e a b = (a - b)

let not_int p = Obj.is_block (Obj.repr p)

type ('component_type, 'component) entity =
  {
    id : int;
    components : ('component_type, 'component) PMap.t;
  }

let new_entity () =
  {
    id = -1;
    components = PMap.empty;
  }

let get_id_opt e =
  if e.id = -1 then None
  else Some e.id

let get_id e =
  if e.id = -1 then invalid_arg "get_id"
  else e.id

let add_component e (comp_type, comp_data) =
  if not_int comp_type then invalid_arg "add_component";
  { e with
    components = PMap.add comp_type comp_data cmp_comp e.components;
  }

let replace_component e comp_type new_val =
  add_component e (comp_type, new_val)

let has_component e comp_type =
  PMap.mem comp_type cmp_comp e.components

let get_component e comp_type =
  PMap.find comp_type cmp_comp e.components

let get_component_opt e comp_type =
  PMap.find_opt comp_type cmp_comp e.components



(** Systems *)

type 'a update =
  | Updated of 'a
  | Unchanged
  | Removed

type ('a, 'b) system = ('a, 'b) entity -> ('a, 'b) world -> delta -> ('a, 'b) entity update * ('a, 'b) entity list
and ('a, 'b) world = {
  id_counter : int;
  entities : (int, ('a, 'b) entity) PMap.t;
  systems : ('a, 'b) system list;
}


let add_system w sys =
  { w with
    systems = sys :: w.systems }



(** World *)

let new_world () =
  {
    id_counter = 0;
    entities = PMap.empty;
    systems = [];
  }


let add_entity w e =
  if e.id = -1 then invalid_arg "add_entity";
  let id = w.id_counter in
  let e = { e with id } in
  { w with
    id_counter = w.id_counter + 1;
    entities = PMap.add id e cmp_e w.entities;
  }


let add_entity_id w e =
  let id = w.id_counter in
  let e = { e with id } in
  { w with
    id_counter = w.id_counter + 1;
    entities = PMap.add id e cmp_e w.entities;
  }, id


let get_entity w id =
  PMap.find id cmp_e w.entities

let get_entity_opt w id =
  PMap.find_opt id cmp_e w.entities

let num_entities w =
  PMap.cardinal w.entities

let add_entities w es =
  List.fold_left add_entity w es

let iter_entities f w =
  PMap.iter (fun e -> f e) w.entities

let fold_entities f w acc =
  PMap.fold (fun e acc -> f e acc) w.entities acc


let replace_entity w id e =
  { w with
    entities = PMap.add id e cmp_e w.entities;
  }


let remove_entity w id e =
  { w with
    entities = PMap.remove id cmp_e w.entities;
  }


let id_mismatch () =
  failwith "world_step: updated entity's id modified"


let world_step w delta =
  PMap.fold
    (fun e w -> let id = e.id in
      List.fold_left (fun w system ->

        let e_up, e_born = system e w delta in
        let w = add_entities w e_born in
        match e_up with
        | Unchanged -> w
        | Removed -> remove_entity w id e
        | Updated e ->
            if (id <> e.id) then id_mismatch ();
            replace_entity w id e

      ) w w.systems
    )
    w.entities
    w
